from setuptools import setup, find_packages

setup(
    name = "mlflow-xethub",
    version = '0.0.0',
    url = "https://github.com/xetdata/Xet-MLflow",
    description = "mlflow[xethub] is a mlflow plugin integrating XetHub with MLflow so that you can use existing MLflow code to track experiments but store artifacts to XetHub.",
    long_description = open('README.md').read(),
    long_description_content_type = "text/markdown",
    author = "XetHub",
    author_email = "contact@xethub.com",
    maintainer = "Kelton Zhang",
    maintainer_email = "kelton@xethub.com",
    license = "BSD-3-Clause",
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
    ],
    keywords = ["ai",
        "collaboration",
        "data-science",
        "developer-tools",
        "git",
        "mlflow",
        "machine-learning",
        "reproducibility",
    ],
    packages=find_packages(),
    # Require MLflow as a dependency of the plugin, so that plugin users can simply install
    # the plugin and then immediately use it with MLflow
    install_requires=["mlflow>=1.27.0", "pyxet>=0.1.0"],
    python_requires='>=3.7',
    entry_points={
        "mlflow.artifact_repository": "xet=mlflow_xet_plugin.xet_artifact:XetHubArtifactRepository",
    },
)
